# app.py
from flask import Flask, render_template, request, redirect, url_for, flash, jsonify, session
import sqlite3
import subprocess
import matplotlib.pyplot as plt
from policy_fixed_point import design_of_experiments
import os
import numpy as np

app = Flask(__name__)
app.secret_key = 'your_secret_key'

# Initialize the SQLite database
def init_db():
    conn = sqlite3.connect('allocations.db', timeout=10)  # Add timeout to the connection
    cursor = conn.cursor()
    cursor.execute('''
        CREATE TABLE IF NOT EXISTS allocations (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            healthcare INTEGER,
            education INTEGER,
            infrastructure INTEGER,
            defense INTEGER,
            environment INTEGER,
            budget INTEGER
        )
    ''')
    cursor.execute('''
        CREATE TABLE IF NOT EXISTS totals (
            id INTEGER PRIMARY KEY,
            healthcare INTEGER DEFAULT 0,
            education INTEGER DEFAULT 0,
            infrastructure INTEGER DEFAULT 0,
            defense INTEGER DEFAULT 0,
            environment INTEGER DEFAULT 0,
            budget INTEGER DEFAULT 0,
            user_count INTEGER DEFAULT 0,
            total_nominal_expenditure INTEGER DEFAULT 0
        )
    ''')
    cursor.execute('''
        CREATE TABLE IF NOT EXISTS votes (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            total_expend REAL,
            vothap_tot REAL,
            healthcare REAL,
            education REAL,
            infrastructure REAL,
            defense REAL,
            environment REAL
        )
    ''')
    cursor.execute('''
        CREATE TABLE IF NOT EXISTS cardinal_vote_totals (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            total_expend REAL,
            vothap_tot REAL,
            healthcare REAL,
            education REAL,
            infrastructure REAL,
            defense REAL,
            environment REAL,
            total_votes INTEGER DEFAULT 0,
            UNIQUE(total_expend, vothap_tot)
        )
    ''')
    # Ensure the votes table has all necessary columns
    try:
        cursor.execute('ALTER TABLE votes ADD COLUMN healthcare REAL')
    except sqlite3.OperationalError:
        pass  # Column already exists
    try:
        cursor.execute('ALTER TABLE votes ADD COLUMN education REAL')
    except sqlite3.OperationalError:
        pass  # Column already exists
    try:
        cursor.execute('ALTER TABLE votes ADD COLUMN infrastructure REAL')
    except sqlite3.OperationalError:
        pass  # Column already exists
    try:
        cursor.execute('ALTER TABLE votes ADD COLUMN defense REAL')
    except sqlite3.OperationalError:
        pass  # Column already exists
    try:
        cursor.execute('ALTER TABLE votes ADD COLUMN environment REAL')
    except sqlite3.OperationalError:
        pass  # Column already exists
    cursor.execute('''
        INSERT OR IGNORE INTO totals (id) VALUES (1)
    ''')
    # Alter the table to add the budget column if it doesn't exist
    try:
        cursor.execute('ALTER TABLE totals ADD COLUMN budget INTEGER DEFAULT 0')
    except sqlite3.OperationalError:
        pass  # Column already exists
    # Alter the table to add the user_count column if it doesn't exist
    try:
        cursor.execute('ALTER TABLE totals ADD COLUMN user_count INTEGER DEFAULT 0')
    except sqlite3.OperationalError:
        pass  # Column already exists
    # Alter the table to add the total_nominal_expenditure column if it doesn't exist
    try:
        cursor.execute('ALTER TABLE totals ADD COLUMN total_nominal_expenditure INTEGER DEFAULT 0')
    except sqlite3.OperationalError:
        pass  # Column already exists
    conn.commit()
    conn.close()

init_db()

def load_effects_table():
    effects_table = []
    with open('effects_array.txt', 'r') as file:
        lines = file.readlines()
        headers = lines[0].strip().split('\t')
        for line in lines[1:]:
            values = line.strip().split('\t')
            effects_table.append({
                'policy': values[0],
                'healthcare': values[1],
                'education': values[2],
                'infrastructure': values[3],
                'defense': values[4],
                'environment': values[5]
            })
    return effects_table

def load_cta():
    cta = {}
    with open('cta.txt', 'r') as file:
        for line in file:
            policy, cost = line.strip().split()
            cta[policy.lower()] = float(cost)
    return cta

def load_pareto_points():
    pareto_points = []
    with open('pareto_frontier.txt', 'r') as file:
        next(file)  # Skip header
        for line in file:
            total_expend, vothap_tot = line.strip().split('\t')
            pareto_points.append({'total_expend': total_expend, 'vothap_tot': vothap_tot})
    return pareto_points

def load_all_points():
    all_points = []
    with open('vothap_results.txt', 'r') as file:
        next(file)  # Skip header
        for line in file:
            exp1, exp2, exp3, exp4, exp5, total_expend, vothap_tot = line.strip().split('\t')
            all_points.append({'total_expend': total_expend, 'vothap_tot': vothap_tot})
    return all_points

def get_expenditures_for_point(total_expend, vothap_tot):
    with open('vothap_results.txt', 'r') as file:
        next(file)  # Skip header
        for line in file:
            exp1, exp2, exp3, exp4, exp5, total_expend_line, vothap_tot_line = line.strip().split('\t')
            if float(total_expend_line) == total_expend and float(vothap_tot_line) == vothap_tot:
                return {
                    'healthcare': exp1,
                    'education': exp2,
                    'infrastructure': exp3,
                    'defense': exp4,
                    'environment': exp5
                }
    return None

@app.route('/')
def index():
    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('SELECT healthcare, education, infrastructure, defense, environment, budget, user_count, total_nominal_expenditure FROM totals WHERE id = 1')
    totals = cursor.fetchone()
    conn.close()
    
    total_points = sum(totals[:-2])
    percentage_healthcare = (totals[0] / total_points) * 100 if total_points > 0 else 0
    percentage_education = (totals[1] / total_points) * 100 if total_points > 0 else 0
    percentage_infrastructure = (totals[2] / total_points) * 100 if total_points > 0 else 0
    percentage_defense = (totals[3] / total_points) * 100 if total_points > 0 else 0
    percentage_environment = (totals[4] / total_points) * 100 if total_points > 0 else 0
    percentage_budget = (totals[5] / total_points) * 100 if total_points > 0 else 0
    
    average_nominal_expenditure = totals[7] / totals[6] if totals[6] > 0 else 0
    
    effects_table = load_effects_table()
    cta = load_cta()
    pareto_points = load_pareto_points() if os.path.exists('pareto_frontier.txt') else None
    all_points = load_all_points() if os.path.exists('vothap_results.txt') else None
    
    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('SELECT AVG(total_expend), AVG(vothap_tot) FROM votes')
    average_pareto_point = cursor.fetchone()
    
    cursor.execute('SELECT DISTINCT total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment FROM votes')
    unique_votes = cursor.fetchall()
    conn.close()
    
    closest_pareto_point = None
    closest_pareto_expenditures = None
    if average_pareto_point[0] is not None and average_pareto_point[1] is not None:
        closest_pareto_point = find_closest_pareto_point(average_pareto_point, pareto_points)
        if closest_pareto_point:
            closest_pareto_expenditures = get_expenditures_for_point(
                float(closest_pareto_point['total_expend']),
                float(closest_pareto_point['vothap_tot'])
            )
    
    # Mark Pareto points
    if (pareto_points):
        pareto_costs, pareto_values = find_pareto_frontier(
            [float(point['total_expend']) for point in pareto_points],
            [float(point['vothap_tot']) for point in pareto_points]
        )
        for point in pareto_points:
            point['is_pareto'] = (float(point['total_expend']) in pareto_costs and float(point['vothap_tot']) in pareto_values)
    
    return render_template('index.html', total_healthcare=totals[0], total_education=totals[1], total_infrastructure=totals[2], total_defense=totals[3], total_environment=totals[4], total_budget=totals[5], effects_table=effects_table, percentage_healthcare=percentage_healthcare, percentage_education=percentage_education, percentage_infrastructure=percentage_infrastructure, percentage_defense=percentage_defense, percentage_environment=percentage_environment, percentage_budget=percentage_budget, grand_total=total_points, cta_healthcare=cta['healthcare'], cta_education=cta['education'], cta_infrastructure=cta['infrastructure'], cta_defense=cta['defense'], cta_environment=cta['environment'], user_count=totals[6], average_nominal_expenditure=average_nominal_expenditure, pareto_points=pareto_points, all_points=all_points, average_pareto_point={'total_expend': average_pareto_point[0], 'vothap_tot': average_pareto_point[1]}, closest_pareto_point=closest_pareto_point, closest_pareto_expenditures=closest_pareto_expenditures, unique_votes=unique_votes, winning_point=session.get('winning_point'))

@app.route('/submit', methods=['POST'])
def submit():
    healthcare = int(request.form['healthcare'])
    education = int(request.form['education'])
    infrastructure = int(request.form['infrastructure'])
    defense = int(request.form['defense'])
    environment = int(request.form['environment'])
    budget = int(request.form['budget'])
    nominal_expenditure = int(request.form['nominal_expenditure'])

    total = healthcare + education + infrastructure + defense + environment + budget

    if total > 100:
        flash('Total allocation cannot exceed 100 points.')
        return redirect(url_for('index'))

    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('''
        UPDATE totals
        SET healthcare = healthcare + ?,
            education = education + ?,
            infrastructure = infrastructure + ?,
            defense = defense + ?,
            environment = environment + ?,
            budget = budget + ?,
            user_count = user_count + 1,
            total_nominal_expenditure = total_nominal_expenditure + ?
        WHERE id = 1
    ''', (healthcare, education, infrastructure, defense, environment, budget, nominal_expenditure))
    conn.commit()
    conn.close()

    flash('Allocation submitted successfully!')
    return redirect(url_for('index'))

def find_pareto_frontier(costs, values):
    sorted_indices = np.argsort(costs)
    pareto_costs = []
    pareto_values = []
    max_value = -np.inf
    for idx in sorted_indices:
        cost = costs[idx]
        value = values[idx]
        if value > max_value:
            pareto_costs.append(cost)
            pareto_values.append(value)
            max_value = value
    return pareto_costs, pareto_values

def write_pareto_underlying_points():
    pareto_points = load_pareto_points()
    with open('pareto_front_underlying_points.txt', 'w') as file:
        file.write("TotalEXPEND\tVOTHAPtot\tHealthcare\tEducation\tInfrastructure\tDefense\tEnvironment\n")
        for point in pareto_points:
            expenditures = get_expenditures_for_point(float(point['total_expend']), float(point['vothap_tot']))
            if expenditures:
                file.write(f"{point['total_expend']}\t{point['vothap_tot']}\t{expenditures['healthcare']}\t{expenditures['education']}\t{expenditures['infrastructure']}\t{expenditures['defense']}\t{expenditures['environment']}\n")

@app.route('/run_pareto', methods=['POST'])
def run_pareto():
    design_of_experiments()
    write_pareto_underlying_points()
    flash('Pareto Frontier analysis completed successfully!')
    return redirect(url_for('index'))

@app.route('/display_db')
def display_db():
    db_contents = fetch_db_contents()
    conn = sqlite3.connect('allocations.db', timeout=10)  # Add timeout to the connection
    cursor = conn.cursor()
    cursor.execute('SELECT user_count, total_nominal_expenditure FROM totals WHERE id = 1')
    result = cursor.fetchone()
    user_count = result[0]
    total_nominal_expenditure = result[1]
    
    cursor.execute('SELECT total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment FROM votes')
    votes = cursor.fetchall()
    
    cursor.execute('SELECT total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment, total_votes FROM cardinal_vote_totals')
    cardinal_votes = cursor.fetchall()
    
    conn.close()
    
    votes = [{'total_expend': vote[0], 'vothap_tot': vote[1], 'healthcare': vote[2], 'education': vote[3], 'infrastructure': vote[4], 'defense': vote[5], 'environment': vote[6]} for vote in votes]
    cardinal_votes = [{'total_expend': vote[0], 'vothap_tot': vote[1], 'healthcare': vote[2], 'education': vote[3], 'infrastructure': vote[4], 'defense': vote[5], 'environment': vote[6], 'total_votes': vote[7]} for vote in cardinal_votes]
    
    return render_template('display_db.html', db_contents=db_contents, user_count=user_count, total_nominal_expenditure=total_nominal_expenditure, votes=votes, cardinal_votes=cardinal_votes, winning_point=session.get('winning_point'))

def find_closest_pareto_point(average_pareto_point, pareto_points):
    min_distance = float('inf')
    closest_point = None
    for point in pareto_points:
        distance = ((float(point['total_expend']) - average_pareto_point[0]) ** 2 + (float(point['vothap_tot']) - average_pareto_point[1]) ** 2) ** 0.5
        if distance < min_distance:
            min_distance = distance
            closest_point = point
    return closest_point

@app.route('/vote', methods=['POST'])
def vote():
    selected_point = request.form['selected_point']
    total_expend, vothap_tot = map(float, selected_point.split('_'))

    underlying_point = get_expenditures_for_point(total_expend, vothap_tot)

    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('''
        INSERT INTO votes (total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment)
        VALUES (?, ?, ?, ?, ?, ?, ?)
    ''', (total_expend, vothap_tot, underlying_point['healthcare'], underlying_point['education'], underlying_point['infrastructure'], underlying_point['defense'], underlying_point['environment']))
    conn.commit()

    cursor.execute('SELECT AVG(total_expend), AVG(vothap_tot) FROM votes')
    average_pareto_point = cursor.fetchone()
    
    cursor.execute('SELECT DISTINCT total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment FROM votes')
    unique_votes = cursor.fetchall()
    conn.close()

    flash('Vote recorded successfully!')

    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('SELECT healthcare, education, infrastructure, defense, environment, budget, user_count, total_nominal_expenditure FROM totals WHERE id = 1')
    totals = cursor.fetchone()
    conn.close()
    
    total_points = sum(totals[:-2])
    percentage_healthcare = (totals[0] / total_points) * 100 if total_points > 0 else 0
    percentage_education = (totals[1] / total_points) * 100 if total_points > 0 else 0
    percentage_infrastructure = (totals[2] / total_points) * 100 if total_points > 0 else 0
    percentage_defense = (totals[3] / total_points) * 100 if total_points > 0 else 0
    percentage_environment = (totals[4] / total_points) * 100 if total_points > 0 else 0
    percentage_budget = (totals[5] / total_points) * 100 if total_points > 0 else 0
    
    average_nominal_expenditure = totals[7] / totals[6] if totals[6] > 0 else 0
    
    effects_table = load_effects_table()
    cta = load_cta()
    pareto_points = load_pareto_points()
    all_points = load_all_points()
    
    closest_pareto_point = None
    closest_pareto_expenditures = None
    if average_pareto_point[0] is not None and average_pareto_point[1] is not None:
        closest_pareto_point = find_closest_pareto_point(average_pareto_point, pareto_points)
        if closest_pareto_point:
            closest_pareto_expenditures = get_expenditures_for_point(
                float(closest_pareto_point['total_expend']),
                float(closest_pareto_point['vothap_tot'])
            )
    
    return render_template('index.html', total_healthcare=totals[0], total_education=totals[1], total_infrastructure=totals[2], total_defense=totals[3], total_environment=totals[4], total_budget=totals[5], effects_table=effects_table, percentage_healthcare=percentage_healthcare, percentage_education=percentage_education, percentage_infrastructure=percentage_infrastructure, percentage_defense=percentage_defense, percentage_environment=percentage_environment, percentage_budget=percentage_budget, grand_total=total_points, cta_healthcare=cta['healthcare'], cta_education=cta['education'], cta_infrastructure=cta['infrastructure'], cta_defense=cta['defense'], cta_environment=cta['environment'], user_count=totals[6], average_nominal_expenditure=average_nominal_expenditure, pareto_points=pareto_points, all_points=all_points, average_pareto_point={'total_expend': average_pareto_point[0], 'vothap_tot': average_pareto_point[1]}, closest_pareto_point=closest_pareto_point, closest_pareto_expenditures=closest_pareto_expenditures, unique_votes=unique_votes)

@app.route('/clear_db', methods=['POST'])
def clear_db():
    conn = sqlite3.connect('allocations.db', timeout=10)  # Add timeout to the connection
    cursor = conn.cursor()
    cursor.execute('''
        UPDATE totals
        SET healthcare = 0,
            education = 0,
            infrastructure = 0,
            defense = 0,
            environment = 0,
            budget = 0,
            user_count = 0,
            total_nominal_expenditure = 0
        WHERE id = 1
    ''')
    cursor.execute('DELETE FROM votes')
    cursor.execute('DELETE FROM cardinal_vote_totals')  # Clear the cardinal_vote_totals table
    conn.commit()
    conn.close()
    
    # Remove the Pareto frontier graph and table
    if os.path.exists('static/pareto_frontier.png'):
        os.remove('static/pareto_frontier.png')
    if os.path.exists('pareto_frontier.txt'):
        os.remove('pareto_frontier.txt')
    
    flash('Database and Pareto frontier cleared successfully!')
    return redirect(url_for('index'))

@app.route('/get_underlying_point')
def get_underlying_point():
    total_expend = float(request.args.get('total_expend'))
    vothap_tot = float(request.args.get('vothap_tot'))
    underlying_point = get_expenditures_for_point(total_expend, vothap_tot)
    return jsonify(underlying_point)

@app.route('/submit_cardinal_vote', methods=['POST'])
def submit_cardinal_vote():
    votes = request.form.getlist('votes[]')
    conn = sqlite3.connect('allocations.db', timeout=10)  # Add timeout to the connection
    cursor = conn.cursor()
    
    cursor.execute('SELECT DISTINCT total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment FROM votes')
    unique_votes = cursor.fetchall()
    
    for i, vote in enumerate(votes):
        if i < len(unique_votes) and int(vote) > 0:
            cursor.execute('''
                INSERT INTO cardinal_vote_totals (total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment, total_votes)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                ON CONFLICT(total_expend, vothap_tot) DO UPDATE SET total_votes = total_votes + excluded.total_votes
            ''', (unique_votes[i][0], unique_votes[i][1], unique_votes[i][2], unique_votes[i][3], unique_votes[i][4], unique_votes[i][5], unique_votes[i][6], int(vote)))
    
    # Ensure the closest Pareto point to average is included in the cardinal_vote_totals table
    cursor.execute('SELECT AVG(total_expend), AVG(vothap_tot) FROM votes')
    average_pareto_point = cursor.fetchone()
    
    closest_pareto_point = find_closest_pareto_point(
        (float(request.form['average_pareto_point_total_expend']), float(request.form['average_pareto_point_vothap_tot'])),
        load_pareto_points()
    )
    if closest_pareto_point:
        closest_pareto_expenditures = get_expenditures_for_point(
            float(closest_pareto_point['total_expend']),
            float(closest_pareto_point['vothap_tot'])
        )
        closest_pareto_votes = int(votes[-1]) if len(votes) > len(unique_votes) else 0
        cursor.execute('''
            INSERT INTO cardinal_vote_totals (total_expend, vothap_tot, healthcare, education, infrastructure, defense, environment, total_votes)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?)
            ON CONFLICT(total_expend, vothap_tot) DO UPDATE SET total_votes = total_votes + excluded.total_votes
        ''', (
            closest_pareto_point['total_expend'],
            closest_pareto_point['vothap_tot'],
            closest_pareto_expenditures['healthcare'],
            closest_pareto_expenditures['education'],
            closest_pareto_expenditures['infrastructure'],
            closest_pareto_expenditures['defense'],
            closest_pareto_expenditures['environment'],
            closest_pareto_votes  # Update vote count for the closest Pareto point
        ))
    
    conn.commit()
    
    # Determine the winning point
    cursor.execute('SELECT total_expend, vothap_tot, MAX(total_votes) FROM cardinal_vote_totals')
    winning_point = cursor.fetchone()
    conn.close()
    
    # Store the winning point in the session
    session['winning_point'] = {'total_expend': winning_point[0], 'vothap_tot': winning_point[1]}
    
    flash('Cardinal vote submitted successfully!')
    
    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('SELECT healthcare, education, infrastructure, defense, environment, budget, user_count, total_nominal_expenditure FROM totals WHERE id = 1')
    totals = cursor.fetchone()
    conn.close()
    
    total_points = sum(totals[:-2])
    percentage_healthcare = (totals[0] / total_points) * 100 if total_points > 0 else 0
    percentage_education = (totals[1] / total_points) * 100 if total_points > 0 else 0
    percentage_infrastructure = (totals[2] / total_points) * 100 if total_points > 0 else 0
    percentage_defense = (totals[3] / total_points) * 100 if total_points > 0 else 0
    percentage_environment = (totals[4] / total_points) * 100 if total_points > 0 else 0
    percentage_budget = (totals[5] / total_points) * 100 if total_points > 0 else 0
    
    average_nominal_expenditure = totals[7] / totals[6] if totals[6] > 0 else 0
    
    effects_table = load_effects_table()
    cta = load_cta()
    pareto_points = load_pareto_points()
    all_points = load_all_points()
    
    closest_pareto_point = None
    closest_pareto_expenditures = None
    if average_pareto_point[0] is not None and average_pareto_point[1] is not None:
        closest_pareto_point = find_closest_pareto_point(average_pareto_point, pareto_points)
        if closest_pareto_point:
            closest_pareto_expenditures = get_expenditures_for_point(
                float(closest_pareto_point['total_expend']),
                float(closest_pareto_point['vothap_tot'])
            )
    
    return render_template('index.html', total_healthcare=totals[0], total_education=totals[1], total_infrastructure=totals[2], total_defense=totals[3], total_environment=totals[4], total_budget=totals[5], effects_table=effects_table, percentage_healthcare=percentage_healthcare, percentage_education=percentage_education, percentage_infrastructure=percentage_infrastructure, percentage_defense=percentage_defense, percentage_environment=percentage_environment, percentage_budget=percentage_budget, grand_total=total_points, cta_healthcare=cta['healthcare'], cta_education=cta['education'], cta_infrastructure=cta['infrastructure'], cta_defense=cta['defense'], cta_environment=cta['environment'], user_count=totals[6], average_nominal_expenditure=average_nominal_expenditure, pareto_points=pareto_points, all_points=all_points, average_pareto_point={'total_expend': average_pareto_point[0], 'vothap_tot': average_pareto_point[1]}, closest_pareto_point=closest_pareto_point, closest_pareto_expenditures=closest_pareto_expenditures, unique_votes=unique_votes, winning_point=session.get('winning_point'))

def fetch_db_contents():
    conn = sqlite3.connect('allocations.db', timeout=10)  # Add timeout to the connection
    cursor = conn.cursor()
    cursor.execute('SELECT healthcare, education, infrastructure, defense, environment, budget, user_count, total_nominal_expenditure FROM totals WHERE id = 1')
    totals = cursor.fetchone()
    conn.close()
    return [
        {'healthcare': totals[0], 'education': totals[1], 'infrastructure': totals[2], 'defense': totals[3], 'environment': totals[4], 'budget': totals[5], 'user_count': totals[6], 'total_nominal_expenditure': totals[7]}
    ]

@app.route('/get_vote_totals')
def get_vote_totals():
    conn = sqlite3.connect('allocations.db', timeout=10)  # Add timeout to the connection
    cursor = conn.cursor()
    cursor.execute('SELECT total_votes FROM cardinal_vote_totals ORDER BY id')
    vote_totals = cursor.fetchall()
    conn.close()
    return jsonify([vote[0] for vote in vote_totals])

if __name__ == '__main__':
    #app.run(debug=True)
    app.run(debug=True, host='127.0.0.1', port=5001)