# ...existing code...

import numpy as np
import matplotlib.pyplot as plt
import sqlite3

def voter_happiness(P, Pb, EFF, IMPmax, CTA, EXPEND, EXPENDnom):
    #Calculate the total priority and the cost per point for each policy
    #Ptot = np.sum(P) + Pb #Total priority (sum of policy priorities and budget priority)
    CPP = CTA/IMPmax #Cost per point for each policy
    # print('CPP = ', CPP)
    EXPENDtot = EXPENDb = np.sum(EXPEND) #Total expenditure on policies
    #EXPENDtot = EXPENDb = 0.0
    IWM = EXPEND/CPP #Improvement with money -- how much improvement is achieved solely through expenditure
    # print('IWM = ', IWM)
    #The voter is most happy if no money is spent on any policy, thus giving IMPmax for the budget
    # beyond that happiness decreases linearly with expenditure
    IMPb = IMPmax - (IMPmax/EXPENDnom)*EXPENDtot #Improvement from budget
    IMP = IWM #initial IMP
    tolerance = 1e-6
    max_iterations = 1000
    for _ in range(max_iterations):
        new_IMP = IWM + np.dot(EFF, IMP)
        new_IMP = np.clip(new_IMP, None, IMPmax)  # Ensure IMP does not exceed 100
        if np.allclose(new_IMP, IMP, atol=tolerance):
            break
        IMP = new_IMP
    # print('Converged IMP = ', IMP)
    VOTHAP = IMP*P #Voter's happiness from each policy
    # print('VOTHAP = ', VOTHAP)
    # print('IMPb = ', IMPb)
    VOTHAPb = IMPb*Pb #Voter's happiness from budget
    # print('VOTHAPb = ', VOTHAPb)
    VOTHAPtot = np.sum(VOTHAP) + VOTHAPb
    # print('VOTHAPtot = ', VOTHAPtot)
    return VOTHAPtot

def find_pareto_frontier(costs, values):
    sorted_indices = np.argsort(costs)
    pareto_costs = []
    pareto_values = []
    max_value = -np.inf
    for idx in sorted_indices:
        cost = costs[idx]
        value = values[idx]
        if value > max_value:
            pareto_costs.append(cost)
            pareto_values.append(value)
            max_value = value
    return pareto_costs, pareto_values

def get_user_input():
    #Manual USER INPUT
    IMPmax = 100.0 #maximum improvement possible for each policy
    #P = np.array([0.2, 0.2, 0.2, 0.2, 0.1]) #Priority of each policy
    #Pb = 0.1 #Priority of budget
    
    # Read EFF from effects_array.txt
    EFF = []
    with open('effects_array.txt', 'r') as file:
        lines = file.readlines()
        for line in lines[1:]:
            values = line.strip().split('\t')[1:]
            EFF.append([float(value) for value in values])
    EFF = np.array(EFF)
    print('EFF = ', EFF)

    # Read CTA from cta.txt
    CTA = []
    with open('cta.txt', 'r') as file:
        for line in file:
            _, cost = line.strip().split()
            CTA.append(float(cost))
    CTA = np.array(CTA)
    print('CTA = ', CTA)

    EXPEND  = np.array([297.96, 112.23, 266.61, 300.0, 400.0]) #Actual expenditure on each policy
    #EXPENDnom = 1000.0 #Nominal expenditure (the expenditure at which voters are neutral)

    # Read values from the database
    conn = sqlite3.connect('allocations.db')
    cursor = conn.cursor()
    cursor.execute('SELECT healthcare, education, infrastructure, defense, environment, budget, user_count, total_nominal_expenditure FROM totals WHERE id = 1')
    totals = cursor.fetchone()
    conn.close()

    P_db = np.array(totals[:6])
    NumUsers = totals[6]
    EXPEND_db = totals[7]

    print('P_db = ', P_db)
    print('NumUsers = ', NumUsers)
    print('EXPEND_db = ', EXPEND_db)
    
    EXPENDnom = EXPEND_db/NumUsers
    Psum = np.sum(P_db)
    P = P_db/Psum
    P = P[:-1]
    Pb = P_db[-1]/Psum
    print('P = ', P)
    print('Pb = ', Pb)
    print('EXPENDnom = ', EXPENDnom)

    return IMPmax, P, Pb, EFF, CTA, EXPEND, EXPENDnom

def design_of_experiments():
    IMPmax, P, Pb, EFF, CTA, EXPEND, EXPENDnom = get_user_input()
    #Design of Experiments
    # Vary the EXPEND values randomly and see how VOTHAPtot changes
    #num_points = 2000
    num_points = 4000
    EXPEND_ranges = [np.random.uniform(0, cta, num_points) for cta in CTA]

    total_expenditures = []
    vothap_totals = []

    
    # Open file to write the table
    with open('vothap_results.txt', 'w') as file:
        # Write headers
        file.write("EXPEND[0]\tEXPEND[1]\tEXPEND[2]\tEXPEND[3]\tEXPEND[4]\tTotalEXPEND\tVOTHAPtot\n")
        print("EXPEND[0]\tEXPEND[1]\tEXPEND[2]\tEXPEND[3]\tEXPEND[4]\tTotalEXPEND\tVOTHAPtot")
    
        for i in range(num_points):
            exp1 = EXPEND_ranges[0][i]
            exp2 = EXPEND_ranges[1][i]
            exp3 = EXPEND_ranges[2][i]
            exp4 = EXPEND_ranges[3][i]
            exp5 = EXPEND_ranges[4][i]
            if(P[0] == 0.0):
                exp1 = 0.0
            if(P[1] == 0.0):
                exp2 = 0.0
            if(P[2] == 0.0):
                exp3 = 0.0
            if(P[3] == 0.0):
                exp4 = 0.0
            if(P[4] == 0.0):
                exp5 = 0.0
            EXPEND_test = np.array([exp1, exp2, exp3, exp4, exp5])
            total_expend = np.sum(EXPEND_test)
            VOTHAPtot = voter_happiness(P, Pb, EFF, IMPmax, CTA, EXPEND_test, EXPENDnom)
            result_line = f'{exp1}\t{exp2}\t{exp3}\t{exp4}\t{exp5}\t{total_expend}\t{VOTHAPtot}\n'
            file.write(result_line)
            print(result_line.strip())
        
            total_expenditures.append(total_expend)
            vothap_totals.append(VOTHAPtot)

    # Find Pareto frontier
    pareto_costs, pareto_values = find_pareto_frontier(total_expenditures, vothap_totals)

    # Write Pareto frontier to a file
    with open('pareto_frontier.txt', 'w') as pareto_file:
        pareto_file.write("TotalEXPEND\tVOTHAPtot\n")
        for cost, value in zip(pareto_costs, pareto_values):
            pareto_file.write(f'{cost}\t{value}\n')

    # Plotting the results -- if you want to see the results in Matplotlib natively
    #plt.scatter(total_expenditures, vothap_totals, label='Total Results')
    #plt.scatter(pareto_costs, pareto_values, color='red', label='Pareto Frontier')
    #plt.xlabel('Total Expenditure')
    #plt.ylabel('Total Voter Happiness')
    #plt.title('Total Expenditure vs Total Voter Happiness')
    #plt.legend()
    #plt.grid(True)
    #plt.show()


#To run a single point
#IMPmax, P, Pb, EFF, CTA, EXPEND, EXPENDnom = get_user_input()
#VOTHAPtot = voter_happiness(P, Pb, EFF, IMPmax, CTA, EXPEND, EXPENDnom)
#print('VOTHAPtot = ', VOTHAPtot)
#To run the design of experiments
#design_of_experiments()
